// $Id: CosEvent_Service.cpp 82011 2008-06-17 13:58:27Z sma $

#include "orbsvcs/CosNamingC.h"
#include "orbsvcs/CosEvent/CEC_EventChannel.h"
#include "orbsvcs/CosEvent/CEC_Default_Factory.h"
#include "orbsvcs/CosEvent/CEC_Event_Loader.h"
#include "ace/Get_Opt.h"
#include "ace/Argv_Type_Converter.h"
#include "ace/OS_main.h"

ACE_RCSID (CosEvent_Service,
           CosEvent_Service,
           "$Id: CosEvent_Service.cpp 82011 2008-06-17 13:58:27Z sma $")

// Check for "--daemon" option and daemonize if present.
namespace
{
  void check_for_daemon (int &argc, ACE_TCHAR* argv[])
  {
    ACE_Get_Opt get_opts (argc, argv, ACE_TEXT("-"));
    get_opts.long_option("daemon");
    int c;

    while ((c = get_opts ()) != -1)
      {
        switch (c)
          {
          case 0:
            {
              // Daemonize, parent exits, child returns here.
              ACE::daemonize();

              // Remove the option from argv []
              // to avoid any confusion that might result.
              for (int i = get_opts.opt_ind (); i != argc; ++i)
                argv [i-1 ] = argv [i];

              // Decrement the value of argc to reflect the removal
              // of the option.
              argc = argc - 1;
              break;
            }
          case '?':
          default:
            // Don't do anything ... much more parsing elsewhere.
            break;
          }
      }
  }
}

int
ACE_TMAIN (int argc, ACE_TCHAR* argv[])
{
  TAO_CEC_Default_Factory::init_svcs ();

  try
    {
      // Copy command line parameter.
      ACE_Argv_Type_Converter command_line(argc, argv);

      // Instead of using -ORBDaemon we check for a "--daemon" flag
      // explicitly.  This allows us to daemonize early, before
      // calling ORB_init.
      check_for_daemon (command_line.get_argc(),
                        command_line.get_TCHAR_argv());

      // Intialize the ORB
      CORBA::ORB_var orb =
        CORBA::ORB_init (command_line.get_argc(), command_line.get_ASCII_argv());

      // Call TAO_CEC_Event_Loader::init (argc, argv) from here.
      TAO_CEC_Event_Loader event_service;

      // To intialise the service
      int result =
        event_service.init (command_line.get_argc(), command_line.get_TCHAR_argv());

      if (result == -1)
        return 1;

      // @@ Run forever... even after the EC is destroyed...
      //    The "Right Way"[tm] to fix this is to modify
      //    TAO_CEC_EventChannel to get a new option.. If the option is
      //    set then we destroy the ORB on EventChannel::destroy(),
      //    otherwise we don't.
      //    The option will not be set by default, but the user can pass
      //    it in the command line, and in this main program with set it
      //    to 1 by default or something like that...
      //    Don't worry about this change yet... Let's get all the changes
      //    in and then we can fix the EC shutdown problem...
      //
      //    As inidicated above, The Typed EC implementation can now be
      //    destroyed by passing -d at the command line and
      //    calling destroy on the typed EC interface.
      //    Calling fini() completes the destruction, although most of
      //    this is done in TAO_CEC_TypedEventChannel::shutdown().
      orb->run ();

#if defined (TAO_HAS_TYPED_EVENT_CHANNEL)
      event_service.fini();
#endif /* TAO_HAS_TYPED_EVENT_CHANNEL */

      // Destroy the ORB
      orb->destroy();
    }
  catch (const CORBA::Exception& ex)
    {
      ex._tao_print_exception (argv[0]);
      return 1;
    }
  return 0;
}
